'''
Most commonly needed functions to plot the data.
'''

import os
import math
import copy

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import proj3d
import mpl_toolkits.axes_grid1
import matplotlib.image
import matplotlib.colors
import matplotlib.cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.ndimage import rotate
import PIL

from .common import (
        vector_plot,
        surface_plot,
        add_rhabdomeres,
        add_line,
        plot_2d_opticflow,
        plot_guidance
        )
from .start_position import start_heatmap, StartposAnalyser
from gonioanalysis.directories import CODE_ROOTDIR
from gonioanalysis.drosom.optic_flow import field_error, field_pvals
from gonioanalysis.coordinates import rotate_vectors
from tk_steroids.routines import extend_keywords
from ..analysers import analyser_classes

plt.rcParams.update({'font.size': 12})

EYE_COLORS = {'right': 'blue', 'left': 'red'}
REPEAT_COLORS = ['green', 'orange', 'pink']


DEFAULT_ELEV = 10
DEFAULT_AZIM = 70
DEFAULT_FIGSIZE = (16,9)
DEFAULT_DPI = 100

def plot_1d_magnitude(manalyser, image_folder=None, i_repeat=None,
        mean_repeats=False, mean_imagefolders=False, mean_eyes=False,
        color_eyes=False, gray_repeats=False, progressive_colors=False,
        show_mean=False, show_std=False,
        show_label=True, milliseconds=False, microns=False,
        phase=False, derivative=False,
        label="EYE-ANGLE-IREPEAT", ax=None):
    '''
    Plots 1D displacement magnitude over time, separately for each eye.
    
    Arguments
    ---------
    manalyser : object
        MAnalyser object instance
    image_folder : string or None
        Image folder to plot the data from. If None (default), plot all image folders.
    mean_repeats : bool
        Wheter to take mean of the repeats or plot each repeat separately
    mean_imagefolders : bool
        If True and image_folder is None (plotting all image folders), takes the mean
        of all image folders.
    mean_eyes : bool
        Wheter to take a mean over the left and right eyes
    label : string
        Label to show. If None, no label. Otherwise
        EYE gets replaced with eye
        ANGLE gets replaced by image folder name
        IREPEAT gets reaplaced by the number of repeat
    
    Returns
        ax
            Matplotlib axes
        traces
            What has been plotted
        N_repeats
            The total number of repeats (independent of i_repeat)
    '''
    
    def get_x_yscaler(mag_rep_i):    
        # FIXME Pixel size and fs should be read from the data
        pixel_size = manalyser.get_pixel_size(image_folder)
        fs = manalyser.get_imaging_frequency(image_folder)
        N = len(mag_rep_i)

        if milliseconds:
            # In milliseconds
            X = 1000* np.linspace(0, N/fs, N)
        else:
            X = np.arange(N)
        
        if microns:
            yscaler = pixel_size
        else:
            yscaler = 1
        
        return X, yscaler
    
    X = None
    yscaler = None

    if ax is None:
        fig, ax = plt.subplots()

    if mean_eyes:
        eyes = [None]
    else:
        eyes = manalyser.eyes

    N_repeats = 0
    traces = []
    


    for eye in eyes:
        magtraces = manalyser.get_magnitude_traces(eye, image_folder=image_folder,
                mean_repeats=mean_repeats, mean_imagefolders=mean_imagefolders,
                _phase=phase, _derivative=derivative)
        
        for angle, repeat_mags in magtraces.items():
            
            if X is None or yscaler is None:
                X, yscaler = get_x_yscaler(repeat_mags[0])
            
            if progressive_colors:
                if gray_repeats:
                    cmap = matplotlib.cm.get_cmap('binary', len(repeat_mags))
                else:
                    cmap = matplotlib.cm.get_cmap('viridis', len(repeat_mags))

            for _i_repeat, mag_rep_i in enumerate(repeat_mags):
                
                N_repeats += 1
                
                if i_repeat is not None and _i_repeat != i_repeat:
                    continue
                
                
                if label:
                    if eye is None:
                        eyename = '+'.join(manalyser.eyes)
                    else:
                        eyename = eye
                    _label = label.replace('EYE', eyename).replace('ANGLE', str(angle)).replace('IREPEAT', str(_i_repeat))
                else:
                    _label = ''
                
                Y = yscaler * mag_rep_i

                if color_eyes:
                    ax.plot(X, Y, label=_label, color=EYE_COLORS.get(eye, 'green'))
                elif progressive_colors:
                    color = cmap(_i_repeat)
                    ax.plot(X, Y, label=_label, color=color)
                else:
                    if gray_repeats:
                        ax.plot(X, Y, label=_label, color='gray')
                    else:
                        ax.plot(X, Y, label=_label)
                
                traces.append(Y)
    
    meantrace = np.mean(traces, axis=0)
    if show_mean:
        ax.plot(X, meantrace, label='mean-of-all', color='black', lw=3)

    if show_std:
        ax.plot(X, meantrace+np.std(traces, axis=0), '--', label='std-of-mean-of-all', color='black', lw=2)
        ax.plot(X, meantrace-np.std(traces, axis=0), '--', color='black', lw=2)
    
    if label and show_label:
        ax.legend(fontsize='xx-small', labelspacing=0.1, ncol=int(len(traces)/10)+1, loc='upper left')    
    

    if milliseconds:
        ax.set_xlabel('Time (ms)')
    else:
        ax.set_xlabel('Frame')
    
    if isinstance(manalyser, analyser_classes['transmittance']):
        ax.set_ylabel('Brightness (counts)')
    else:
        if microns:
            ax.set_ylabel('Displacement magnitude (µm)')
        else:
            ax.set_ylabel('Displacement magnitude (pixels)')


    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    return ax, traces, N_repeats


def plot_xy_trajectory(manalysers, wanted_imagefolders=None, i_repeat=None,
        mean_repeats=True,
        ax=None):
    '''
    A (x,y) 2D plot of the movement trajectory where time is encoded
    in color.

    manalysers : list
    wanted_imagefolders : dict
        {specimen_name: [image_folder1, ...]}
    i_repeat : int
    mean_repeats : bool
    '''
    if isinstance(manalysers[0], StartposAnalyser):
        xys = []
        for manalyser in manalysers:
            if wanted_imagefolders:
                image_folders = wanted_imagefolders[manalyser.name]
            else:
                image_folders = manalyser.list_imagefolders()
            xys.append(start_heatmap(manalyser, image_folders, ax=ax))
        
        return ax, xys

    if ax is None:
        fig, ax = plt.subplots()

    xys = []

    
    for manalyser in manalysers:

        if wanted_imagefolders:
            image_folders = wanted_imagefolders[manalyser.name]
        else:
            image_folders = manalyser.list_imagefolders()
        
        for image_folder in image_folders:

            movement_data = manalyser.get_movements_from_folder(image_folder)

            for eye, movements in movement_data.items():
                
                N_repeats = len(movements)
                X = [[-x for x in movements[i]['x']] for i in range(N_repeats) if not (
                    (i_repeat is not None) and i_repeat != i)]

                Y = [movements[i]['y'] for i in range(N_repeats) if not (
                    (i_repeat is not None) and i_repeat != i)]

                if mean_repeats:
                    X = [np.mean(X, axis=0)]
                    Y = [np.mean(Y, axis=0)]

                for x, y in zip(X, Y):  
                    N = len(x) 
                    cmap = matplotlib.cm.get_cmap('tab20', N)
                    for i_point in range(1, N):
                        ax.plot([x[i_point-1], x[i_point]], [y[i_point-1], y[i_point]], color=cmap((i_point-1)/(N-1)))
                
                    ax.scatter(x[0], y[0], color='black')
                    ax.scatter(x[-1], y[-1], color='gray')
                
                    xys.append([x, y])


    # Colormap
    if xys:
        if getattr(ax, 'xy_colorbar', None) is None:
            time = [i for i in range(N)]
            sm = matplotlib.cm.ScalarMappable(cmap=cmap)
            sm.set_array(time)

            fig = ax.get_figure()
            ax.xy_colorbar = fig.colorbar(sm, ticks=time, boundaries=time, ax=ax, orientation='horizontal')
            ax.xy_colorbar.set_label('Frame')
        else:
            pass

    ax.set_xlabel('Displacement in X (pixels)')
    ax.set_ylabel('Displacement in Y (pixels)')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    ax.yaxis.set_ticks_position('left')
    ax.xaxis.set_ticks_position('bottom')
    ax.set_aspect('equal', adjustable='box')

    return ax, xys


def plot_magnitude_probability(manalysers, wanted_imagefolders=None, ax=None,
        mean_repeats=False, mean_imagefolders=False,
        microns=True, milliseconds=True, xpoints=100):
    '''
    With many analysers and image folders, calculate probablity of the magnitude
    response (2D histogram).

    Arguments
    ---------
    manalysers : list of objs
        Analyser objects
    wanted_imagefolders : None or dict
        Keys are analyser names, items lists of image folder names
        to incldue. If None, use all image folders.
    '''
    print(wanted_imagefolders)

    for logaritmic in [False, True]:

        N_real = None

        all_magtraces = []

        for manalyser in manalysers:
            for eye in manalyser.eyes:
                
                if wanted_imagefolders:
                    image_folders = wanted_imagefolders[manalyser.name]
                else:
                    image_folders = manalyser.list_imagefolders()
                
                for image_folder in image_folders:
                    magtraces = manalyser.get_magnitude_traces(eye, image_folder=image_folder,
                            mean_repeats=mean_repeats, mean_imagefolders=mean_imagefolders)

                    magtraces = list(magtraces.values())
                    
                    if len(magtraces) == 0:
                        continue

                    magtraces = magtraces[0]
                    
                    if N_real is None:
                        N_real = len(magtraces[0])
                    elif N_real != len(magtraces[0]):
                        raise ValueError('All magtrace not the same length')

                    # Interpolate
                    magtraces = [np.interp(np.linspace(0,1,xpoints), np.linspace(0,1,len(magtrace)), magtrace) for magtrace in magtraces]
                    
                    all_magtraces.extend(magtraces)
        
        all_magtraces = np.array(all_magtraces, dtype=np.float)

        limits = (min(0, np.min(all_magtraces)), np.max(all_magtraces)*1.5)
        tmax = len(all_magtraces[0]) * (N_real/xpoints)
        

        #if ax is None:
        fig, ax = plt.subplots()
        
        if microns:
            pixel_size = manalyser.get_pixel_size(image_folder)
            limits = [z*pixel_size for z in limits]
        ax.set_ylabel('Displacement (µm)')
        
        if milliseconds:
            tmax *= 1000 / manalyser.get_imaging_frequency(image_folder) 
            ax.set_xlabel('Time (ms)')
        else:
            ax.set_xlabel('i_frame')
        
        prob = []

        for points_t in all_magtraces.T:
            hist = np.histogram(points_t, range=limits, bins=25)[0]
            prob.append(hist/np.sum(hist))
        
        if logaritmic:
            norm = matplotlib.colors.LogNorm()
        else:
            norm = None

        im = ax.imshow(np.array(prob).T, origin='lower', cmap='binary',
                norm=norm,
                extent=[0, tmax,limits[0], limits[1]],
                aspect='auto')# interpolation='lanczos')
        
        divider = make_axes_locatable(ax)
        cax = divider.append_axes('right', size='5%', pad=0.05) 
        ax.figure.colorbar(im, cax)

    return ax, prob


def _set_analyser_attributes(analyser, skip_none=True, raise_errors=False, **kwargs):
    '''
    Sets attributes to manalyser.
    Raises AttributeError if the analyser does not have the attribute beforehand.

    Returns a dictionary of the original parameters
    '''
    for key, value in kwargs:
        if value is not None or skip_none == False:
            if hasattr(analyser, key):
                setattr(analyser, key, value)
            elif raise_errors:
                raise AttributeError('{} has no attribute {} prior setting'.format(anlayser, key))


def plot_2d_vectormap(manalyser,
        ax=None):
    '''
    Plots a 2-dimensional vector map.

    Arguments
    ----------
    manalyser : object
        Instance of MAnalyser class or MAverager class (having get2DVectors method)
    ax : object
        Matplotlib axes object
    '''
    
    if not ax:
        fig, ax = plt.subplots(figsize=(8,16))

    hmin = 99999
    hmax = -99999
    vmin = 99999
    vmax = -99999

    for color, eye in zip(['red', 'blue'], ['left', 'right']):
        angles, X, Y = manalyser.get_2d_vectors(eye,
                mirror_horizontal=False, mirror_pitch=False)
        for angle, x, y in zip(angles, X, Y):
           
            horizontal, pitch = angle
            
            hmin = min(hmin, horizontal)
            hmax = max(hmax, horizontal)
            vmin = min(vmin, pitch)
            vmax = max(vmax, pitch)

            # If magnitude too small, its too unreliable to judge the orientation so skip over
            #movement_magnitude = math.sqrt(x**2 + y**2)
            #if movement_magnitude < 2:
            #    continue 
            # Scale all vectors to the same length
            #scaler = math.sqrt(x**2 + y**2) / 5 #/ 3
            #scaler = 0
            #if scaler != 0:
            #    x /= scaler
            #    y /= scaler /2.4    # FIXME
            scaler = np.linalg.norm([x, y]) / 8
            x /= scaler
            y /= scaler

            #ar = matplotlib.patches.Arrow(horizontal, pitch, xc, yc)
            ar = matplotlib.patches.FancyArrowPatch((horizontal, pitch),
                    (horizontal-x, pitch+y), mutation_scale=10,
                    color=color, picker=True)
            #fig.canvas.mpl_connect('pick_event', self.on_pick)
            ax.add_patch(ar)

            ax.scatter(horizontal, pitch, marker='x', color='gray')
    
    ax.set_xlabel('Horizontal rotation (degrees)')
    ax.set_ylabel('Vertical rotation (degrees)')
    
    hmin = -60
    hmax = 40
    vmin = -90
    vmax = 120

    ax.set_xlim(hmin-10, hmax+10)
    ax.set_ylim(vmin-10, vmax+10)
    ax.set_aspect('equal', adjustable='box')
    
    for key in ['top', 'right']:
        ax.spines[key].set_visible(False)


@extend_keywords(vector_plot)
def plot_3d_vectormap(
        manalyser, arrow_rotations = [0],
        image_folder=None,
        rhabdomeres=False, repeats_separately=False, vertical_hardborder=False,
        merge_distance=None,
        elev=None, azim=None,
        pitch_rot=None, roll_rot=None, yaw_rot=None,
        animation=None, animation_type=None, animation_variable=None, i_frame=0,
        ax=None, **kwargs):
    '''
    Plot a 3D vectormap, where the arrows point movement (MAnalyser, FAnalyser) or
    feature (OAnalyser) directions.
    
    Arguments
    ---------
    manalyser : object
        Analyser object
    arrow_rotations : list of int
        Rotation of arrows in the plane of the arrows (ie. radially).
    image_folder : None or str
        The vector to highlight in orange
    rhabdomeres : bool
        If True, draw rhabdomere pattern where the arrows are.
    repeats_separately : bool
        If True, and repeat data exists, draw each repeat
        with it's own arrow.
    vertical_hardborder : bool
        For MAverager, interpolate dorsal and ventral separetly
    merge_distance : None or float
        If not None, merge (mean) vectors that are closer than the given
        distance
    elev, azim : float or None
        Plot elevation and azim
    animation : None
    animation_type : None
    animation_variable : None
    i_frame : int
    ax : object
        Maptlotlib ax object. If not specified, creates a new figure.
    kwargs : dict
        Variable keyword arguments are passed to vector_plot function.

    Returns
    -------
    ax : object
        Matplotlib Axes object
    vectors : list of objects
        List of arrow artist drawn on the figure.
    '''
    colors = EYE_COLORS
    
    # Work out MAnalyser type
    manalyser_type = 'MAnalyser'

    if manalyser.__class__.__name__ == 'OAnalyser' and len(arrow_rotations) == 1 and arrow_rotations[0] == 0:
        manalyser_type = 'OAnalyser'
    elif manalyser.__class__.__name__ == 'MAverager':
        if all([partanalyser.__class__.__name__ == 'OAnalyser' for partanalyser in manalyser.manalysers]):
            manalyser_type = 'OAnalyser'
    elif manalyser.__class__.__name__ == 'FAnalyser':
        colors = ['darkviolet']*5
        if animation_type != 'rotate_arrows':
            i_frame = 0
        
    _set_analyser_attributes({'pitch_rot': pitch_rot,
        'roll_rot': roll_rot, 'yaw_rot': yaw_rot})


    if manalyser_type == 'OAnalyser':
        colors = REPEAT_COLORS
        i_frame = 0
        
        # OAnalyser specific for Drosophila; Assuming that R3-R6 line is
        # analysed, let's also draw the line from R3 to R1.
        if arrow_rotations[0] == 0 and len(arrow_rotations) == 1:
            arrow_rotations.append(29)

    if ax is None:
        fig = plt.figure(figsize=kwargs.pop('figsize', DEFAULT_FIGSIZE), dpi=kwargs.pop('dpi', DEFAULT_DPI))
        ax = fig.add_subplot(111, projection='3d')
    
    vectors = {}

    original_rotation = manalyser.vector_rotation

    if animation_type == 'rotate_plot':
        elev, azim = animation_variable

    if azim is not None and elev is not None:
        camerapos = (elev, azim)
    else:
        camerapos = None
    
    # For OAnalyser, when rhabdomeres is set True,
    # plot the rhabdomeres also
    if manalyser_type == 'OAnalyser' and rhabdomeres:
        manalyser.vector_rotation = 0
        for eye in manalyser.eyes:

            vectors_3d = manalyser.get_3d_vectors(
                    eye, correct_level=True,
                    repeats_separately=repeats_separately,
                    strict=True, vertical_hardborder=vertical_hardborder,
                    merge_distance=merge_distance)

            if eye == 'left':
                mirror_lr = True
            else:
                mirror_lr = False

            for point, vector in zip(*vectors_3d):
                add_rhabdomeres(ax, *point, *vector,
                        mirror_lr=mirror_lr, mirror_bf='auto',
                        camerapos=camerapos,
                        resolution=9, edgecolor=None, facecolor='gray')

    for i_rotation, rotation in enumerate(arrow_rotations):

        for eye in manalyser.eyes:
            if isinstance(colors, dict):
                colr = colors[eye]
            elif isinstance(colors, list):
                colr = colors[i_rotation]
            
            # Set arrow/vector rotation
            if rotation is not None or rotation != 0:
                manalyser.vector_rotation = rotation
            
            vectors_3d = manalyser.get_3d_vectors(
                    eye, correct_level=True,
                    repeats_separately=repeats_separately,
                    strict=True, vertical_hardborder=vertical_hardborder,
                    merge_distance=merge_distance)
           
            # Dirty hack to color the selected imagefolder in orange
            if image_folder is not None:
                hlvec = manalyser.get_3d_vectors(
                        eye, image_folder=image_folder,
                        correct_level=True,
                        repeats_separately=repeats_separately,
                        strict=True,
                        vertical_hardborder=vertical_hardborder,
                        merge_distance=merge_distance)
                if len(hlvec[0]) > 0:
                    x,y,z = hlvec[0][0]
                    # Remove the highlighted from vectors
                    points = vectors_3d[0].tolist()
                    vecs = vectors_3d[1].tolist()
                    for rmindex, (xx,yy,zz) in enumerate(points):
                        if xx==x and yy==y and zz==z:
                            break
                    points.pop(rmindex)
                    vecs.pop(rmindex)
                    vectors_3d = [np.array(points), np.array(vecs)]

                    vector_plot(ax, *hlvec, color="orange",**kwargs)
               

            if manalyser_type == 'OAnalyser' and rhabdomeres:
                
                for point, vector in zip(*vectors_3d):
                    add_line(ax, *point, *vector, color=REPEAT_COLORS[i_rotation])
            else:
                vector_plot(ax, *vectors_3d, color=colr,
                        **kwargs
                        )
               
            vectors[eye] = vectors_3d
           
    manalyser.vector_rotation = original_rotation

    if hasattr(ax, 'set_xlim3d'): 
        ax.set_xlim3d((-1,1))
        ax.set_ylim3d((-1,1))
        ax.set_zlim3d((-1,1))
        ax.set_box_aspect((1, 1, 1)) 
    
    if azim is None and elev is None:
        ax.view_init(elev=DEFAULT_ELEV, azim=DEFAULT_AZIM)
    else:
        ax.view_init(elev=elev, azim=azim)

    return ax, vectors


def plot_3d_magnitudemap(
        analyser,
        elev=None, azim=None,
        ax=None):
    '''Plot 3d heatmap showing the size of movement

    Arguments
    ---------
    analyser : object
    image_folder : None or str
        
    '''
    
    MAGS = {}

    vmin = np.inf
    vmax = -np.inf

    for eye in analyser.eyes:
        vectors_3d = analyser.get_3d_vectors(
                eye, correct_level=True, repeats_separately=False,
                    strict=True, vertical_hardborder=True,
                    normalize_length=None)

        points, vectors = vectors_3d

        magnitudes = [np.linalg.norm(
            p-v) for p,v in zip(points, vectors)]
        magnitudes = [points, magnitudes]

        MAGS[eye] = magnitudes
        
        if len(magnitudes[1]) > 0:
            vmin = np.min([vmin, np.min(magnitudes[1])])
            vmax = np.max([vmax, np.max(magnitudes[1])])

    print(vmin)
    print(vmax)

    for eye in analyser.eyes:
        mgs = [(m-vmin)/(vmax+vmin) for m in MAGS[eye][1]]
        MAGS[eye][1] = np.array(mgs)

    for eye in analyser.eyes:
        m = surface_plot(ax, *MAGS[eye])
         
    ax.set_xlim3d((-1,1))
    ax.set_ylim3d((-1,1))
    ax.set_zlim3d((-1,1))
    ax.set_box_aspect((1, 1, 1)) 
 
    if azim is None and elev is None:
        ax.view_init(elev=DEFAULT_ELEV, azim=DEFAULT_AZIM)
    else:
        ax.view_init(elev=elev, azim=azim)

    return ax, MAGS
    

def plot_3d_differencemap(manalyser1, manalyser2, ax=None, stats_map=False,
        elev=DEFAULT_ELEV, azim=DEFAULT_AZIM, colinear=True, direction=False,
        colorbar=True, colorbar_text=True, colorbar_ax=None, reverse_errors=False,
        colorbar_text_positions=[[1.1,0.95,'left', 'top'],[1.1,0.5,'left', 'center'],[1.1,0.05,'left', 'bottom']],
        i_frame=0, arrow_rotations=[0], pitch_rot=None, yaw_rot=None, roll_rot=None,
        hide_axes=False, hide_text=False, guidance=False, **kwargs):
    '''
    Plots 3d heatmap presenting the diffrerence in the vector orientations
    for two analyser objects, by putting the get_3d_vectors of both analysers
    to field_error and making a surface plot.
    
    Notes:
        - Errors (differences) are calculated at manalyser1's points.
        - arrow_rotations, pitch_rot only affects manalyser2

    manalyser1, manalyser2 : object
        Analyser objects to plot the difference with 3d vectors
    ax : object or None
        Matplotlib Axes object
    stats_map : bool
        If true, plot p-vals.
    colorbar : bool
        Whether to add the colors explaining colorbar
    colorbar_text: bool
        Wheter to add the text annotations on the colorbar
    colorbar_ax : object
        Optional Axes where to put the colorbar
    arrow_rotations : list
        Arrow rotations, for the second manalyser
    i_frame : int
        Neglected here
    hide_axes : bool
    hide_text : bool
    guidance : bool
    '''
    
    if ax is None:
        fig = plt.figure(figsize=DEFAULT_FIGSIZE)
        ax = fig.add_subplot(111, projection='3d')
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
     
    if arrow_rotations:
        original_rotation = manalyser2.vector_rotation
        manalyser2.vector_rotation = arrow_rotations[0]
    
    if pitch_rot is not None:
        manalyser2.pitch_rot = pitch_rot
    if yaw_rot is not None:
        manalyser2.yaw_rot = yaw_rot
    if roll_rot is not None:
        manalyser2.roll_rot = roll_rot
    
    if guidance:
        plot_guidance(ax, camerapos=(ax.elev, ax.azim), hide_text=hide_text)
    
    if hide_axes:
        ax.set_axis_off()

    if hide_text:
        ax.axes.xaxis.set_ticklabels([])
        ax.axes.yaxis.set_ticklabels([])
        ax.axes.zaxis.set_ticklabels([]) 
        ax.set_xlabel('')
        ax.set_ylabel('')
        ax.set_zlabel('')      
    else:   
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
    

    all_errors = []
    for eye in manalyser1.eyes:
        vectors = []
        points = []

        for manalyser in [manalyser1, manalyser2]:
            
            vectors_3d = manalyser.get_3d_vectors(eye, correct_level=True,
                repeats_separately=stats_map,
                strict=True, vertical_hardborder=True)
 
            points.append(vectors_3d[0])
            vectors.append(vectors_3d[1])
       
        # Errors at the points[0]
        if stats_map:
            points, errors = field_pvals(points[0], vectors[0], points[1], vectors[1], colinear=colinear)
        else:
            # regular difference map
            errors = field_error(points[0], vectors[0], points[1], vectors[1],
                    colinear=colinear, direction=direction)
        
        if reverse_errors:
            errors = 1-errors

        all_errors.append(errors)
        
        if eye=='left':
            all_phi_points = [np.linspace(math.pi/2, 3*math.pi/2, 50)]
        else:
            all_phi_points = [np.linspace(-math.pi/2, math.pi/2, 50)]
            #all_phi_points = [np.linspace(0, math.pi/2, 25), np.linspace(3*math.pi/2, 2*math.pi,25)]

        if direction:
            colormap = 'own-diverge'
        else:
            colormap = 'own'

        print('{} eye, mean error {}'.format(eye, np.mean(errors)))

        for phi_points in all_phi_points:
            m = surface_plot(ax, points[0], errors, phi_points=phi_points,
                    colormap=colormap)
    
    errors = np.concatenate(all_errors)

    ax.view_init(elev=elev, azim=azim)

    # COLORBAR
    cax = getattr(ax, 'differencemap_colorbar_ax', None)
    colorbar_obj = getattr(ax, 'differencemap_colorbar', None)
    if colorbar and (cax is None or colorbar_obj is None): 
        
        if colorbar_ax is None:
            cbox = ax.get_position()
            cbox.x1 -= abs(cbox.x1 - cbox.x0)/10
            cbox.x0 += abs((cbox.x1 - cbox.x0))/1.1
            cbox.y0 -= 0.18*abs(cbox.y1-cbox.y0)
            cax = ax.figure.add_axes(cbox)
        else:
            cax = colorbar_ax
        
        ax.differencemap_colorbar_ax = cax
        ax.differencemap_colorbar = plt.colorbar(m, cax)
    
        # COLORBAR INFO TEXT
        if colorbar_text:
            #text_x = 1+0.1
            #ha = 'left'
            if direction:
                cax.text(colorbar_text_positions[0][0], colorbar_text_positions[0][1],
                        'Counterclockwise +90',
                        ha=colorbar_text_positions[0][2], va=colorbar_text_positions[0][3],
                        transform=cax.transAxes)
                cax.text(colorbar_text_positions[1][0], colorbar_text_positions[1][1],
                        'Perpendicular',
                        ha=colorbar_text_positions[1][2], va=colorbar_text_positions[1][3],
                        transform=cax.transAxes)
                cax.text(colorbar_text_positions[2][0], colorbar_text_positions[2][1], 
                        'Clockwise -90',
                        ha=colorbar_text_positions[2][2], va=colorbar_text_positions[2][3],
                        transform=cax.transAxes)
            elif colinear:
                cax.text(colorbar_text_positions[0][0], colorbar_text_positions[0][1],
                        'Collinear',
                        ha=colorbar_text_positions[0][2], va=colorbar_text_positions[0][3],
                        transform=cax.transAxes)
                cax.text(colorbar_text_positions[2][0], colorbar_text_positions[2][1], 
                        'Perpendicular',
                        ha=colorbar_text_positions[2][2], va=colorbar_text_positions[2][3],
                        transform=cax.transAxes)
            else:
                cax.text(colorbar_text_positions[0][0], colorbar_text_positions[0][1],
                        'Matching',
                        ha=colorbar_text_positions[0][2], va=colorbar_text_positions[0][3],
                        transform=cax.transAxes)
                cax.text(colorbar_text_positions[1][0], colorbar_text_positions[1][1],
                        'Perpendicular',
                        ha=colorbar_text_positions[1][2], va=colorbar_text_positions[1][3],
                        transform=cax.transAxes)
                cax.text(colorbar_text_positions[2][0], colorbar_text_positions[2][1], 
                        'Opposing',
                        ha=colorbar_text_positions[2][2], va=colorbar_text_positions[2][3],
                        transform=cax.transAxes)
            
            cax.set_axis_off()
    elif colorbar is False and colorbar_obj is not None:
        ax.differencemap_colorbar.remove()
        ax.differencemap_colorbar = None

    if arrow_rotations:
        manalyser2.vector_rotation = original_rotation
    
    ax.set_xlim3d((-1,1))
    ax.set_ylim3d((-1,1))
    ax.set_zlim3d((-1,1))
    ax.set_box_aspect((1, 1, 1)) 
 
    return ax, errors


def compare_3d_vectormaps(manalyser1, manalyser2, axes=None,
        illustrate=True, total_error=True, compact=False,
        animation=None, animation_type=None, animation_variable=None,
        optimal_ranges=None, pulsation_length=1, biphasic=False,
        kwargs1={}, kwargs2={}, kwargsD={}, **kwargs):
    '''
    Calls get 3d vectors for both analysers.
    Arrow rotation option only affects manalyser2
   
    Note! If total_error is true, sets attributes pupil_compare_errors
        and pupil_compare_rotations to the last ax in axes

    manalyser1,manalyser2 : objects
        Analyser objects
    axes : list of objects
        List of Matplotlib Axes objects. Expected base length == 3 which
        is modified by options orientation, total_error and compact.
    
    illustrate : bool
        Wheter to plot the illustrative plot. Increases the axes requirement by 1
    total_error: bool
        Whether to plot the total error plot Increases the axes requirement by 1
    compact : bool
        Join vectorplots in one. Decreases axes requirement by 1

    animation : list
        For total error x-axis limits
    kwargs1,kwargs2 : dict
        List of keyword arguments to pass to `plot_3d_vectormap`
    kwargsD : dict
        List of keywords arguments to pass to `plot_3d_differencemap`
    
    Returns [axes]
    '''


    optimal=False

    if axes is None:
        fig = plt.figure(figsize=(DEFAULT_FIGSIZE[0]*2,DEFAULT_FIGSIZE[1]*2))
        axes = []

        N_plots = 3
        
        if compact:
            N_plots -= 1 
        if total_error:
            N_plots += 1
        if illustrate:
            N_plots += 1

        n_rows = 1
        n_cols = N_plots

        for i_plot in range(N_plots):
            axes.append(fig.add_subplot(n_rows, n_cols, i_plot+1, projection='3d'))



    kwargs1 = kwargs.copy()
    kwargs2 = kwargs.copy()
    
    if animation_type == 'rotate_arrows':
        kwargs2['arrow_rotations'] = [animation_variable]
        kwargsD['colorbar_text_positions'] = [[0.5,1.03,'center', 'bottom'],
                [1.1,0.5,'left', 'center'],
                [0.5,-0.03,'center', 'top']]
        
        if manalyser1.__class__.__name__ == 'FAnalyser':
            manalyser1.constant_points = True

    elif animation_type in ['pitch_rot', 'roll_rot', 'yaw_rot']:
        kwargs2[animation_type] = animation_variable
        kwargsD['colinear'] = False
    elif animation_type == 'rotate_plot':
        kwargs1['elev'] = animation_variable[0]
        kwargs1['azim'] = animation_variable[1]
        kwargs2['elev'] = animation_variable[0]
        kwargs2['azim'] = animation_variable[1]

    # set 10 deg pitch for flow
    for manalyser in [manalyser1, manalyser2]:
        if manalyser.manalysers[0].__class__.__name__ == 'FAnalyser' and animation_type != 'pitch_rot':
            if manalyser == manalyser1:
                kwargs1['pitch_rot'] = 10
            if manalyser == manalyser2:
                kwargs2['pitch_rot'] = 10
            
            manalyser.pitch_rot = 10

    if manalyser2.__class__.__name__ == 'MAverage' and manalyser2.manalysers[0].__class__.__name__ == 'OAnalyser':
        kwargs2['arrows'] = False
    
    iax = 0
    plot_3d_vectormap(manalyser1, animation_type=animation_type, ax=axes[iax], **kwargs1)
    
    if not compact:
        iax += 1
    plot_3d_vectormap(manalyser2, animation_type=animation_type, ax=axes[iax], **kwargs2)
    

    if biphasic:
        # Difference with the slow phase
        iax += 1
        kwargsDr = kwargsD.copy()
        kwargsDr['colorbar'] = False
        daxr, reverse_errors = plot_3d_differencemap(manalyser1, manalyser2,
                ax=axes[iax], reverse_errors=True, **kwargsDr, **kwargs2)
    
    
    iax += 1
    dax, errors = plot_3d_differencemap(manalyser1, manalyser2,
            ax=axes[iax], **kwargsD, **kwargs2)
    
    if illustrate:

        # Check if we are at the optimal
        if optimal_ranges:
            for optimal_range in optimal_ranges:
                if optimal_range[0] < animation_variable < optimal_range[1]:
                    optimal=True
                    continue
        
        iax += 1
        ax = axes[iax]

        if animation_type == 'rotate_arrows':
            
            upscale = 4
            image = matplotlib.image.imread(os.path.join(CODE_ROOTDIR, 'images', 'dpp.tif'))
            image = PIL.Image.fromarray(image)
            ow,oh = image.size
            image = image.resize((int(ow*upscale), int(oh*upscale)), PIL.Image.NEAREST)
            image = np.array(image)

            # Vector p0,rp1 to point the current axis rotation
            R6R3line = 40
            rot = -math.radians( R6R3line + animation_variable ) 
            p0 = [int(image.shape[0]/2), int(image.shape[1]/2)]
            p1 = np.array([p0[0],0])/2
            rp1 = np.array([p1[0]*math.cos(rot)-p1[1]*math.sin(rot), p1[0]*math.sin(rot)+p1[1]*math.cos(rot)])
            
            # Optimal
            orot = -math.radians( R6R3line + np.mean(optimal_ranges[0][0:2]))
            orp1 = np.array([p1[0]*math.cos(orot)-p1[1]*math.sin(orot), p1[0]*math.sin(orot)+p1[1]*math.cos(orot)])
            
            # Make image pulsate
            sx, sy = (0,0)
            r= (0,0)
            if manalyser1.manalysers[0].__class__.__name__ == 'MAnalyser':
                r = image.shape
                r = [int(pr) for pr in [r[0]/20, r[1]/20, r[0]-r[0]/10, r[1]-r[1]/10]]

                # FIXME 0.6 is the low value of the pulsation
                sx = int(upscale * (orp1[0]) * (pulsation_length-0.8) / 10)
                sy = int(upscale * (orp1[1]) * (pulsation_length-0.8) / 10)
                
                #if not optimal:
                #    sx = 0
                #    sy = 0

                print('Animation pulsataion sx {} sy {}, imshape {}'.format(sx,sy, image.shape))
        
                #image[r[1]+sy:r[1]+r[3]+sy, r[0]+sx:r[0]+r[2]+sx] = image[r[1]:r[1]+r[3], r[0]:r[0]+r[2]]
                image = image[r[1]-sy:r[1]+r[3]-sy, r[0]-sx:r[0]+r[2]-sx]
            ax.imshow(image, cmap='gray')
           
            # R3-R6 dotted white line
            if manalyser1.manalysers[0].__class__.__name__ == 'FAnalyser':
                rot36 = -math.radians( R6R3line )
                rp1_36 = np.array([p1[0]*math.cos(rot36)-p1[1]*math.sin(rot36), p1[0]*math.sin(rot36)+p1[1]*math.cos(rot36)])
                ax.axline((p0[0]-r[0], p0[1]-r[1]), (p0[0]+rp1_36[0]-r[0], p0[1]+rp1_36[1]-r[1]), ls='--', color='white', lw=0.5)

            ax.axline((p0[0]-r[0], p0[1]-r[1]), (p0[0]+rp1[0]-r[0], p0[1]+rp1[1]-r[1]), color=REPEAT_COLORS[0])
            
            # Rhabdomere locations, dpp.tiff specific
            rhabdomere_locs = [(x*upscale+sx-r[0],y*upscale+sy-r[1]) for x,y in [(74,60),(68,79),(58,101),(80,94),(96,87),(100,66),(85,74)]]
            for i_rhabdomere, loc in enumerate(rhabdomere_locs):
                ax.text(*loc, 'R'+str(i_rhabdomere+1), color=(0.2,0.2,0.2), ha='center', va='center', fontsize=10)
            

        elif animation_type == 'pitch_rot':
            
            image = matplotlib.image.imread(os.path.join(CODE_ROOTDIR, 'images', 'from_mikko_annotated.png'))
            ax.imshow(rotate(image, animation_variable, mode='nearest', reshape=False)) 
            plot_2d_opticflow(ax, 'side')

        elif animation_type == 'yaw_rot':
            image = matplotlib.image.imread(os.path.join(CODE_ROOTDIR, 'images', 'rotation_yaw.png'))
            ax.imshow(rotate(image, animation_variable, mode='nearest', reshape=False))
            plot_2d_opticflow(ax, 'side')

        elif animation_type == 'roll_rot':
            image = matplotlib.image.imread(os.path.join(CODE_ROOTDIR, 'images', 'rotation_roll.png'))
            ax.imshow(rotate(image, animation_variable, mode='nearest', reshape=False))
            plot_2d_opticflow(ax, 'outofplane')


        if optimal_ranges:
            for optimal_range in optimal_ranges:
                if optimal_range[0] < animation_variable < optimal_range[1]:
                    rect = matplotlib.patches.Rectangle((0,0), 1, 1, transform=ax.transAxes, fill=False,
                            color='yellow', linewidth=8)
                    
                    ax.add_patch(rect)
                    ax.text(0.5, 0.9, optimal_range[2], ha='center', va='top', color='gold', transform=ax.transAxes, fontsize=12)
                    optimal=True
                    continue
   
        if manalyser1.manalysers[0].__class__.__name__ == 'FAnalyser' and manalyser2.manalysers[0].__class__.__name__ == 'OAnalyser':
            

            if getattr(axes[0], 'extra_illustrate_ax', None) is None:
                
                tmp_ax = axes[0].figure.add_subplot(3,4,8)
                tmp_ax.set_axis_off()
                cbox = tmp_ax.get_position()
                w = abs(cbox.x1 - cbox.x0)
                cbox.x0 += 0.25*w
                cbox.x1 += 0.25*w
                axes[0].extra_illustrate_ax = axes[0].figure.add_axes(cbox)
                axes[0].extra_illustrate_ax.set_frame_on(False)
                axes[0].extra_illustrate_ax.set_axis_off()
            else:
                axes[0].extra_illustrate_ax.clear()
                axes[0].extra_illustrate_ax.set_axis_off()
                axes[0].extra_illustrate_ax.set_frame_on(False)

            image = matplotlib.image.imread(os.path.join(CODE_ROOTDIR, 'images', 'from_mikko_annotated.png'))
            axes[0].extra_illustrate_ax.imshow(rotate(image, manalyser1.pitch_rot, mode='nearest', reshape=False))

            rect = matplotlib.patches.Rectangle((0.9+0.05,0), 0.02, 1, transform=axes[0].extra_illustrate_ax.transAxes, fill=True,
                    color='yellow', linewidth=1)
            
            axes[0].extra_illustrate_ax.add_patch(rect)
            # Optic flow arrows
            arrows = [np.array((1.2,ik))*len(image) for ik in np.arange(0.1,0.91,0.1)]
            for x, y in arrows:
                axes[0].extra_illustrate_ax.arrow(x, y, -0.1*len(image)*pulsation_length/2., 0, width=0.01*len(image), color='darkviolet')
        

    if total_error:
        iax += 1

    

        ax = axes[iax]
        if getattr(ax, 'pupil_compare_errors', None) is None:
            ax.pupil_compare_errors = []
            ax.pupil_compare_rotations = []

        ax.pupil_compare_errors.append(np.mean(errors))
        ax.pupil_compare_rotations.append( animation_variable )

        ax.plot( ax.pupil_compare_rotations, 1-np.array(ax.pupil_compare_errors), color='black',
                label='Fast phase')
        ax.scatter( ax.pupil_compare_rotations[-1], 1-ax.pupil_compare_errors[-1], color='black' )
       

        print('Minimum and maximum errors so far: {} (min, at angle {}), {} (max, at angle {})'.format(
            np.min(ax.pupil_compare_errors), ax.pupil_compare_rotations[np.argmin(ax.pupil_compare_errors)],
            np.max(ax.pupil_compare_errors), ax.pupil_compare_rotations[np.argmax(ax.pupil_compare_errors)]))

        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        
        ax.set_xlabel('Degrees')
        ax.set_ylabel('Mean error')

        if animation is not None:
            ax.set_xlim(np.min(animation), np.max(animation))
            ax.set_ylim(0,1)
            ax.set_yticks([0, 0.5, 1])
        
        if optimal:
            formatting = '{:.1f}'
        else:
            formatting = '{:.0f}'

        if animation_type == 'rotate_arrows':
            text = 'Rotation from R3-R6 line\n{} degrees'
        elif animation_type in ['pitch_rot', 'yaw_rot', 'roll_rot']:
            text = 'Head tilt {} degrees'
        else:
            text = 'Animation variable {}'
        
        if animation_variable is not None:
            text = text.format(formatting).format(float(animation_variable))
            ax.text(0.1, 1, text, transform=ax.transAxes, va='bottom', ha='left',fontsize=12)
        
        if animation:
            if np.min(animation) < -100 and np.max(animation) > 100:
                ax.set_xticks([-90,0,90])
                ax.set_xticklabels(['-90$^\circ$', '0$^\circ$','90$^\circ$'])   
            else:
                ax.set_xticks([-45, 0, 45])
                ax.set_xticklabels(['-45$^\circ$', '0$^\circ$','45$^\circ$'])   
            
        if biphasic:
            if getattr(ax, 'pupil_compare_reverse_errors', None) is None:
                ax.pupil_compare_reverse_errors = []

            ax.pupil_compare_reverse_errors.append(np.mean(reverse_errors))

            ax.plot( ax.pupil_compare_rotations, 1-np.array(ax.pupil_compare_reverse_errors), color='gray',
                    label='Slower phase')
            ax.scatter( ax.pupil_compare_rotations[-1], 1-ax.pupil_compare_reverse_errors[-1], color='gray' )
    
            ax.legend(loc=(0.39,1.2))


    return [axes]



def compare_3d_vectormaps_compact(*args, **kwargs):
    '''
    Wrapper for compare_3d_vectormaps but
        compact=True, total_error=False, illustrate=False
    '''
    return compare_3d_vectormaps(*args, compact=True, total_error=False, illustrate=False, **kwargs)



def compare_3d_vectormaps_manyviews(*args, axes=None,
        column_titles=['Microsaccades', 'Rhabdomere orientation', 'Difference', 'Mean microsaccade'],
        row_titles=['Dorsal\nview', 'Anterior\nview', 'Ventral\nview'],
        **kwargs):
    '''
    Just with different views rendered
    
    First axes gets attributes .orientation_ax and .error_ax
    '''
    

    views = [[50,90], [0,90], [-50,90]]
    rows = len(views)
    cols = 4
    
    biphasic = False
    if kwargs.get('animation_type', None) in ['pitch_rot', 'yaw_rot', 'roll_rot']:
        biphasic = True
        cols += 1
        column_titles.insert(3, 'Difference2')
    
        if kwargs.get('animation_type', None) in ['pitch_rot', 'yaw_rot', 'roll_rot']:
            column_titles[2] = 'Difference\n with slower phase'
            column_titles[3] = 'Difference\n with fast phase'
            column_titles[-1] = ''

        plt.subplots_adjust(left=0.05, bottom=0.04, right=0.9, top=0.94, wspace=0.05, hspace=0.05)
    else:
        plt.subplots_adjust(left=0.1, bottom=0.04, right=0.9, top=0.95, wspace=0.05, hspace=0.05)

    if axes is None:
        fig = plt.figure(figsize=DEFAULT_FIGSIZE,dpi=300)
        axes = []
        

        for i_view in range(rows):
            for column in range(cols-1):
                axes.append(fig.add_subplot(rows,cols,column+1+i_view*cols, projection='3d'))
        
        if getattr(kwargs, 'illustrate_ax', None) in [True, None]:
            # FIXME Very hacky way to move the subplot right. For some reason
            # when moved this way the plot also gets smaller?
            axes[0].illustrate_ax = fig.add_subplot(rows,cols,cols)
            cbox = axes[0].illustrate_ax.get_position()
            w = abs(cbox.x1 - cbox.x0)
            h = abs(cbox.y1 - cbox.y0)
            cbox.x0 -= w/8
            cbox.x1 += w / 3.3 +w/8
            cbox.y1 += h / 3.3
            
            if kwargs.get('animation_type', None) == 'pitch_rot':
                w = abs(cbox.x1 - cbox.x0)
                h = abs(cbox.y1 - cbox.y0)
                cbox.x0 -= w/5
                cbox.x1 += w/5
                cbox.y0 -= w/5
                cbox.y1 += w/5

            axes[0].illustrate_ax.set_position(cbox)
            axes[0].illustrate_ax.cbox = cbox

        if getattr(kwargs, 'total_error', None) in [True, None]:
            ax = fig.add_subplot(rows,cols,3*cols)
            ax_pos = ax.get_position()
            ax_pos = [ax_pos.x0+0.02, ax_pos.y0-0.04, ax_pos.width+0.022, ax_pos.height+0.02]
            ax.remove()
            ax = fig.add_axes(ax_pos)
            axes[0].error_ax = ax

        tmp_ax = fig.add_subplot(rows, cols, 2*cols)
        tmp_ax.set_axis_off()
        cbox = tmp_ax.get_position()
        
        cbox.x1 -= abs(cbox.x1 - cbox.x0)/1.1
        w = abs(cbox.x1 - cbox.x0)
        cbox.x0 -= 2*w
        cbox.x1 -= 2*w
        axes[0].colorbar_ax = ax.figure.add_axes(cbox)
    
    # Clear axes that are attributes of the axes[0]; These won't
    # otherwise get cleared for the animation/video
    
    if getattr(axes[0], 'error_ax', None) is not None:
        axes[0].error_ax.clear()
    
    # Set custom column titles
    for i_manalyser in range(2):
        manalyser = args[i_manalyser]
        if manalyser.manalysers[0].__class__.__name__ == 'FAnalyser' or manalyser.__class__.__name__ == 'FAnalyser':
            if '\n' in ''.join(column_titles):
                column_titles[i_manalyser] = 'Optic flow\n'
            else:
                column_titles[i_manalyser] = 'Optic flow'

        if manalyser.manalysers[0].__class__.__name__ == 'MAnalyser' and manalyser.manalysers[0].receptive_fields == True:
            column_titles[i_manalyser] = 'Biphasic receptive field\nmovement directions'


    if args[0].manalysers[0].__class__.__name__ == 'FAnalyser':
        column_titles[cols-1] = 'Mean optic flow axis'

    if getattr(axes[0], 'illustrate_ax', None) is not None:
        axes[0].illustrate_ax.clear()
        axes[0].illustrate_ax.set_title(column_titles[-1], color=REPEAT_COLORS[0])
        #axes[0].illustrate_ax.text(0.5,1, column_titles[-1], transform=axes[0].illustrate_ax.transAxes, ha='center', va='bottom')        
        axes[0].illustrate_ax.set_frame_on(False)
        axes[0].illustrate_ax.set_axis_off()

    # Add column titles
    for title, ax in zip(column_titles[:-1], axes[0:cols-1]):
        ax.set_title(title)
    
    # Add row titles for the views
    for title, ax in zip(row_titles, axes[::cols-1]):
        if biphasic:
            ax.text2D(-0.1, 0.5, title.replace('\n', ' '), transform=ax.transAxes, va='center', ha='center', rotation=90)
        else:
            ax.text2D(-0.375, 0.5, title, transform=ax.transAxes, va='center')

    for ax in axes:
        ax.set_axis_off()
    
    naxes = cols -1

    for i in range(3):
        viewargs = copy.deepcopy(kwargs)
        viewargs['elev'] = views[i][0]
        viewargs['azim'] = views[i][1]
        
        if i == 0:
            compare_3d_vectormaps(axes=axes[i*naxes:(i+1)*naxes]+[axes[0].illustrate_ax, axes[0].error_ax],
                    biphasic=biphasic,
                    kwargsD={'colorbar': True, 'colorbar_ax': axes[0].colorbar_ax},
                    illustrate=True, total_error=True,
                    *args, **viewargs)
        else:
            compare_3d_vectormaps(axes=axes[i*naxes:(i+1)*naxes]+[axes[0].illustrate_ax, axes[0].error_ax],
                    biphasic=biphasic, illustrate=False, total_error=False,
                    kwargsD={'colorbar': False},
                    *args, **viewargs)
    
    for ax in axes:
        ax.dist = 6

    axes[0].illustrate_ax.set_position(axes[0].illustrate_ax.cbox)
    return [axes]



#!/usr/bin/env python3
'''Command line interface for GonioAnalysis motion measurement.
'''

import sys
import os
import argparse

import matplotlib.pyplot as plt

from gonioanalysis.drosom.analyser_commands import (
        ANALYSER_CMDS,
        DUALANALYSER_CMDS,
        MULTIANALYSER_CMDS,
        )
from gonioanalysis.droso import DrosoSelect
# Import analysers
from gonioanalysis.drosom.analysing import MAverager
from gonioanalysis.drosom.analysers import analyser_classes

# Avoid importing tkinter bits if not needed
if '--tk_waiting_window' in sys.argv:
    from gonioanalysis.tkgui.widgets import WaitingWindow



Analysers = analyser_classes

analyses = {**ANALYSER_CMDS, **DUALANALYSER_CMDS, **MULTIANALYSER_CMDS}



def parse_key_value_pairs(string, splitchar=','):
    return {opt.split('=')[0]: opt.split('=')[1] for opt in string.split(splitchar)}


def convert_kwarg(key, value):
    '''Take good guesses to converte kwargs to right type for analyser_cmds.

    Originally all items are strings.
    '''
    if key in ['elev', 'azim', 'dpi']:
        value = float(value)
    elif value.lower()  == 'true':
        value = True
    elif value.lower() == 'false':
        value = False
    elif key == 'image_folder':
        return value
    elif ',' in value:
        value = value.split(',')
        if all([val.replace('.', '', 1).removeprefix('-').isdigit() for val in value]):
            value = [float(val) for val in value]
    return value


def main(custom_args=None):
    
    if custom_args is None:
        custom_args = sys.argv[1:]
    
    parser = argparse.ArgumentParser(description=__doc__)
    
    # DATA INPUT ARGUMENTS
    parser.add_argument('-D', '--data_directory', nargs='+',
            help='Data directory')

    parser.add_argument('-S', '--specimens', nargs='+',
            help=('Comma separeted list of specimen names.'
                ' Separate groups by space when averaging is on.'
                ' If needed, wanted image folders can be specified with semicolons'
                ' for example, specimen1;imf1;imf2:specimen2;imf1'
                ' (does not work with groups). If image folders contain commas, use'
                ' semicolons, use colons (:) to separate specimens'))
    

    # SETTINGS OR ACTIONS FOR ANALYSERS
    parser.add_argument('-a', '--average', action='store_true',
            help='Average and interpolate the results over the specimens')
    
    parser.add_argument('-t', '--type', nargs='+',
            help='Analyser type, either "motion" or "orientation". Space separate gor groups')
   
    parser.add_argument(
            '--analyser-options', nargs='+',
            help='Extra arguments to ui options. Space separate for groups')

    parser.add_argument('-r', '--reselect-rois', action='store_true',
            help='Reselect ROIs')
 
    parser.add_argument('-R', '--recalculate-movements', action='store_true',
            help='Recalculate with Movemeter')

    parser.add_argument('--active-analysis', nargs='?', default='',
            help="Name of the analyser's active analysis (nothing to do with the --analysis/-A option)")

    # OTHER SETTINGS
    parser.add_argument('--tk_waiting_window', action='store_true',
            help='(internal) Launches a tkinter waiting window')
    parser.add_argument('--dont-show', action='store_true',
            help='Skips showing the plots')
    parser.add_argument('--savefig', action='store_true',
            help='Saves the matplotlib figure if the analysis it produces')

    # Different analyses for separate specimens

    parser.add_argument('-A', '--analysis', nargs=1,
            choices=analyses.keys(),
            help='The performed analysis or action')
   
    parser.add_argument('--analysis-options', nargs='+',
            help='Keyword arguments to the analysis function')

    # Other settings
    parser.add_argument('-o', '--output',
            help='Output filename for export analysis options')


    args = parser.parse_args(custom_args)
    

    if args.tk_waiting_window:
        waiting_window = WaitingWindow('terminal.py', 'When ready, this window closes.')


    # Getting the data directory
    # --------------------------
    if args.data_directory:
        print('Using data directory {}'.format(args.data_directory[0]))
        
        data_directories = args.data_directory
    else:
        # If all FAnalysers, data directoy not needed
        if isinstance(args.type, list) and all([t=='flow' for t in args.type]):
            data_directories = []
        else:
            data_directories = input('Data directory >> ')
    
    # Check that the data directory exists
    for directory in data_directories:
        if not os.path.isdir(directory):
            raise ValueError("{} is not a directory".format(directory))

    # {specimen name : [image_folder_1, ...]}
    wanted_imagefolders = {}

    # Getting the specimens
    # ---------------------
    directory_groups = []
    if args.specimens:
        
        for i_group, group in enumerate(args.specimens):
            print('Using specimens {}'.format(group))
            
            # For flow analysers give fake folders
            if isinstance(args.type, list) and args.type[i_group] == 'flow':
                directory_groups.append( ['none/none'] )
                continue
                
            if group == 'none':
                directory_groups.append(None)
                continue

            if ':' in group:
                specimen_separator = ':'
            else:
                specimen_separator = ','

            if ';' in group:
                
                for specimen in group.split(specimen_separator):
                    if ';' in specimen:
                        splitted = specimen.split(';')
                        wanted_imagefolders[splitted[0]] = splitted[1:]
                
                # Remove specified imagefolders
                group = ','.join([z.split(';')[0] for z in group.split(specimen_separator)])
            
            
            # dont commit me
            group = group.replace(':', ',')
            
            
            directories = []
            for directory in data_directories:
                selector = DrosoSelect(datadir=directory)
                directories.extend( selector.parse_specimens(group) )
            directory_groups.append(directories)
    else:
        selector = DrosoSelect(datadir=data_directories[0])
        directories = selector.ask_user()
     
            
    # Setting up analysers
    # ---------------------
    
    if not args.type:
        args.type = ['motion' for i in directory_groups]

    analyser_groups = []
    
    for i_group, directories in enumerate(directory_groups):

        analysers = []
        Analyser = Analysers[args.type[i_group]]
        
        print('Using {}'.format(Analyser.__name__))

        if directories is None:
            analysers.append(Analyser(None, None))
        else:

            for directory in directories: 
                
                path, folder_name = os.path.split(directory)
                analyser = Analyser(path, folder_name) 
                
                if args.active_analysis:
                    analyser.active_analysis = args.active_analysis
                
                analysers.append(analyser)
         
        if args.analyser_options:
            print(f'  Setting options {args.analyser_options[i_group]}')
            for analyser in analysers:
                # Special empty ones to sign that for analysers in i_group
                # we don't want to set any options
                if args.analyser_options[i_group].lower() in ['na', 'none']:
                    continue
                # Parse analyser options
                opts = parse_key_value_pairs(args.analyser_options[i_group]) 
                analyser.set_ui_options(opts)

        # Ask ROIs if not selected
        for analyser in analysers:
            if analyser.are_rois_selected() == False or args.reselect_rois:
                analyser.select_ROIs()

        # Analyse movements if not analysed, othewise load these
        for analyser in analysers:
            if analyser.is_measured() == False or args.recalculate_movements:
                analyser.measure_movement(eye='left')
                analyser.measure_movement(eye='right')
            analyser.load_analysed_movements()
        
        
        if args.average:
            
            if len(analysers) >= 2:

                avg_analyser = MAverager(analysers)
                avg_analyser.setInterpolationSteps(5,5)
                
                analysers = avg_analyser
            else:
                analysers = analysers[0]
        else:
            if len(analysers) == 1:
                analysers = analysers[0]

        analyser_groups.append(analysers)
    

    function = analyses[args.analysis[0]]
    
    print(analyser_groups)
    
    kwargs = {}
    if wanted_imagefolders:
        kwargs['wanted_imagefolders'] = wanted_imagefolders
    if args.output:
        kwargs['save_fn'] = args.output
    if args.analysis_options:
        for aopts in args.analysis_options:
            for key, value in parse_key_value_pairs(aopts, splitchar=';;;').items():
                kwargs[key] = convert_kwarg(key, value)

    if function in MULTIANALYSER_CMDS.values():
        for analysers in analyser_groups:
            function(analysers, **kwargs)
    elif args.average or function in DUALANALYSER_CMDS.values():
        function(*analyser_groups, **kwargs)
    else:
        for analysers in analyser_groups:
            if not isinstance(analysers, list):
                analysers = [analysers]
            for analyser in analysers:
                function(analyser, **kwargs)

    if args.tk_waiting_window:
        waiting_window.close()
    
    if args.savefig:
        if args.output:
            plt.savefig(args.output, transparent=True)
        else:
            plt.savefig('figure.png', transparent=True)
    else:
        if not args.dont_show:
            plt.show()


if __name__ == "__main__":
    main()





from .analysing import MAnalyser, MAverager
from .orientation_analysis import OAnalyser
from .optic_flow import FAnalyser
from .transmittance_analysis import TAnalyser, RelativeTAnalyser
from .startpos_analysis import StartposAnalyser


analyser_classes = {
        'orientation': OAnalyser,
        'motion': MAnalyser,
        'flow': FAnalyser,
        'transmittance': TAnalyser,
        'relative-transmittance': RelativeTAnalyser,
        'startposition': StartposAnalyser}



'''
Functions to create plots using matplotlib out of
XAnalyser objects.
'''

import os
import json
import warnings

import tifffile
import matplotlib.pyplot as plt
import matplotlib.patches

from numpy import sign, mean
import numpy as np
import scipy.interpolate

from gonioanalysis.directories import PROCESSING_TEMPDIR, ANALYSES_SAVEDIR


def plot_1d_overlap(xanalysers):
    '''
    Plot binocular overlap so that the x-axis is the pitch angle (vertical angle)
    and the y-axis is the width of binocular overlap.

    xanalysers      List of XAnalyser objects

    Returns the created figure.
    '''
    X = []
    Y = []
    flies = []
    
    figure = plt.figure()

    for xanalyser in xanalysers:

        X.append([])
        Y.append([])

        for marking in sorted(xanalyser.get_overlaps(), key=lambda x: x['pitch']):
            Y[-1].append(abs(marking['horizontal_left'] - marking['horizontal_right']))
            X[-1].append(marking['pitch'])
        
        flies.append(xanalyser.fly)
    

    # Interpolate the data; Required for the mean traces
    # ----------------------------------------------------
    intp_step = 1

    XXs_span = np.arange(int(np.min(np.min(X))/intp_step)*intp_step, int(np.max(np.max(X))/intp_step)*intp_step, intp_step)

    XX = []
    YY = []

    for fly, x, y in zip(flies, X,Y):
        xx = np.arange(int(np.min(x)/intp_step)*intp_step, int(np.max(x)/intp_step)*intp_step, intp_step)
        yy = np.interp(xx, x, y)
        plt.scatter(xx, yy, s=5)
        XX.append(xx)
        YY.append(yy)
    

    # Mean trace
    # ----------------
    mean_YY = []

    for x in XXs_span:
        yys_to_average = []
        
        for yy, xx in zip(YY, XX):
            try:
                index = list(xx).index(x)
            except:
                continue
            
            yys_to_average.append(yy[index])
        
        if yys_to_average:
            mean_YY.append(np.mean(yys_to_average))
        else:
            mean_YY.append(0)
    

    plt.plot(XXs_span, mean_YY, linewidth=2)
    plt.xlabel('Vertical angle (degrees)')
    plt.ylabel('Binocular overlap (degrees)')

    return figure


def plot_matrix_overlap(xanalysers):
    '''
    Plot the binocular overlap in a kind of "matrix representation" by discreting
    the continues data further, and plotting colored squared.
    '''

    def _plotMatrix(matrix, newfig=False, subplot=111):
        '''
        Temporarily hidden here. Needs commeting.
        '''
        matrix_height, matrix_width = matrix.shape
        
        if newfig == True:
            plt.figure()
        plt.subplot(subplot)

        plt.imshow(matrix, cmap='coolwarm', interpolation='none',
                extent=(hor_range[0], hor_range[1], ver_range[1], ver_range[0]),
                aspect='auto')

        ax = plt.gca();
        ax.set_xticks(np.arange(hor_range[0]+hor_step, hor_range[1]+hor_step, hor_step), minor=True)
        ax.set_yticks(np.arange(ver_range[0]+ver_step, ver_range[1]+ver_step, ver_step), minor=True) 
        ax.grid(which='minor', color='black', linestyle='-', linewidth=1)


    hor_step = 1
    ver_step = 4
    hor_range = (-40, 40)
    ver_range = (-90, 90)
    
    matrices = []

    matrix_width = int((hor_range[1]-hor_range[0])/hor_step)
    matrix_height = int((ver_range[1]-ver_range[0])/ver_step)

    for xanalyser in xanalysers:
        fly = xanalyser.fly
        markings = xanalyser.get_overlaps()

        X = []
        Y = []
        midpoints =  [] 
        
        # FIXME or not?
        # The following code for the matrix overlap plot goes quite verbose.
        # Maybe it could be made mode readable? At least commenting whats
        # going on in each section. Seems to work never the less.

        for marking in markings:
            
            mid = marking['horizontal_middle']
            marking['horizontal_left'] - mid
            
            row = []

            marking['horizontals'].sort()
            
            if marking['horizontal_right'] - marking['horizontal_left'] < hor_step:
                row = [0]*matrix_width
                row[int(matrix_width/2)] = 2

            else:
                for angle in range(hor_range[0], hor_range[1], hor_step):
                    if angle < marking['horizontal_left']:
                        row.append(0)
                    elif marking['horizontal_left'] <= angle <= marking['horizontal_right']:
                        row.append(1)
                    elif marking['horizontal_right'] < angle:
                        row.append(0)
            
            midpoints.append( (marking['horizontal_middle']) / (int(hor_range[1]-hor_range[0])/2) )

            if len(row) != matrix_width:
                print(row)
                print(marking['horizontal_left'])
                print(marking['horizontal_right'])
                raise UserWarning('Row length {} but matrix width {}'.format(len(row), matrix_width))
      

            X.append(row)
            Y.append(marking['pitch'])
        
        matrix = np.zeros( (int((ver_range[1]-ver_range[0])/ver_step), int((hor_range[1]-hor_range[0])/hor_step)) )
        matrix_i_midpoint = int(matrix.shape[0] / 2)
        for j, pitch in enumerate(range(*ver_range, ver_step)):

            indices = [y for y in Y if pitch-ver_step/2 <= y <= pitch+ver_step/2]
            indices = [Y.index(y) for y in indices]
            
            for index in indices:
                i_midpoint = int((midpoints[index])*int(matrix_width/2))
                shift = -1*(i_midpoint)
                if shift >= 0:
                    matrix[j][shift:] += np.asarray(X[index])[0:matrix_width-shift]
                elif shift < 0:
                    matrix[j][0:matrix_width+shift] += np.asarray(X[index])[-shift:]
                    
                matrix = np.round(matrix)
                matrix = np.clip(matrix, 0, 1) 
        matrices.append(matrix)
        
    avg_matrix = matrices[0] / len(matrices)
    for i in range(1, len(matrices)):
        avg_matrix += matrices[i] / len(matrices)
    
    
    matrix = np.round(avg_matrix)
    
    _plotMatrix(avg_matrix, newfig=True)

    for j in range(0, avg_matrix.shape[0]):
        row_max = np.max(avg_matrix[j])
        if row_max > np.min(avg_matrix[j]):
            avg_matrix[j] /= row_max
    
    figure = plt.figure() 
    
    #FIXME
    '''
    for i, matrix in enumerate(matrices):
        for j in range(0, matrix.shape[0]):
            if not np.any(matrix[j]):
                matrix[j] += 0.5

        _plotMatrix(matrix, subplot=int('{}{}{}'.format(3,round(len(matrices)/3),i+1)))
    '''
    _plotMatrix(avg_matrix, newfig=True)

    # Phiuw, we're out.
    return figure


def plot_experiment_illustration(xanalyser):
    '''
    Plot a video of hor the fly was rotated while simultaneously reconstructing
    the matrix plot.
    '''
    
    dpi = 300

    savepath = os.path.join(ANALYSES_SAVEDIR, 'binocular_overlap_illustration', xanalyser.fly)
    os.makedirs(savepath, exist_ok=True)

    hor_step = 1
    ver_step = 4
    hor_range = (-40, 40)
    ver_range = (-90, 90)
    
    fig = plt.figure(figsize=(8, 4), dpi=dpi)
    ax = fig.add_subplot(121)
    #ax.set_axis_off()
    ax.set_xlim(*hor_range)
    ax.set_ylim(*ver_range)
    
    imax = fig.add_subplot(122)
    imax.set_axis_off()
    
    markings = sorted(xanalyser.get_overlaps(), key=lambda x: x['pitch'])

    # Variables to keep track of the steps    
    visited_pitches = []
    i_image = 0
    direction = True
    
    antenna_level = xanalyser.get_antenna_level()

    for pitch, horizontals_images in xanalyser.get_data():
        

        visited_horizontals = []

        if visited_pitches and abs(pitch - visited_pitches[-1]) < ver_step:
            continue
        
        print('Pitch {}'.format(pitch))

        # Find the marking of this pitch
        for marking in markings:
            if marking['pitch'] == pitch - antenna_level:
                break
        
        hm = marking['horizontal_middle']
        hl = marking['horizontal_left']
        hr = marking['horizontal_right']
    
        if visited_pitches:
            if pitch > visited_pitches[-1]:
                y = visited_pitches[-1] + ver_step/2 - 0.1
                h = (pitch + ver_step/2) - y
            else:
                print('naan')
                y2 = visited_pitches[-1] - ver_step/2
                y1 = pitch - ver_step/2
                y = y1
                h = abs(y2 - y1)
        else:
            y = pitch - ver_step/2
            h = ver_step

        if direction:
            horizontals_images.reverse()
            direction = False
        else:
            direction = True

        for horizontal, image_fn in horizontals_images:
            
            if visited_horizontals and abs(horizontal - visited_horizontals[-1]) < hor_step:
                continue
           
            if not hor_range[0] <= horizontal <= hor_range[1]:
                continue

            # Decide color of drawing
            if hl <= horizontal <= hr:
                color = "purple"
            elif horizontal < hl:
                color = "red"
            elif horizontal > hr:
                color = 'blue'
            
            if visited_horizontals:
                if horizontal > visited_horizontals[-1]:
                    x = visited_horizontals[-1] + hor_step/2
                    w = (horizontal + hor_step/2) - x
                else:
                    x2 = visited_horizontals[-1] - hor_step/2
                    x1 = horizontal - hor_step/2
                    x = x1
                    w = x2 - x1
            else:
                x = horizontal - hor_step/2
                w = hor_step
            
            
            rect = matplotlib.patches.Rectangle((-(x+w-hm), -(y+h-antenna_level)), w, h, color=color)
            ax.add_patch(rect)

            image = tifffile.imread(image_fn)
            image = np.clip(image, 0, np.percentile(image, 95))
            image = np.rot90(image) / np.max(image)

            try:
                imshow_obj.set_data(image)
            except UnboundLocalError:
                imshow_obj = imax.imshow(image, cmap='gray')
            
                
            fig.savefig(os.path.join(savepath, 'image_{:08d}'.format(i_image)), dpi=dpi)
            i_image += 1
            
            
            
            visited_horizontals.append(horizontal)
        
        visited_pitches.append(pitch)



    

# FIXME or remove?
'''
def plot2DOverlap(xanalysers):
    
    Xl = []
    Yl = []
    Xr = []
    Yr = []
    Xm = []
    Ym = []



    for marking in self.overlap_markings:
        
        mid = marking['horizontal_middle']
        #mid = 0
        Xl.append(marking['horizontal_left']-mid)
        Xr.append(marking['horizontal_right']-mid)
        Xm.append(marking['horizontal_middle']-mid)
        Yl.append(marking['pitch'])
        Yr.append(marking['pitch'])
        Ym.append(marking['pitch'])
    
    plt.scatter(Xl, Yl, color='blue')
    plt.scatter(Xr, Yr, color='red')
    plt.scatter(Xm, Ym, color='yellow')


    plt.show()

   
def plotFancy2DOverlap():

    X,Y,C = [[],[],[]]
    for marking in self.overlap_markings:

        mid = marking['horizontal_middle']
        
        for i in range(len(marking['horizontals'])):
            
            pitch = marking['pitch']
            horizontal = marking['horizontals'][i]
            L = min(marking['horizontal_left'], marking['horizontal_right'])
            R = max(marking['horizontal_left'], marking['horizontal_right'])
            
            if L < horizontal < R:
                C.append(2)
            else:
                C.append(1)
            
            X.append(horizontal-mid)
            Y.append(pitch)

    f_int = scipy.interpolate.interp2d(X, Y, C, fill_value=1)
    
    X = np.linspace(np.min(X), np.max(X), 100)
    Y = np.linspace(np.min(Y), np.max(Y), 100)
    C = f_int(X, Y)
    
    X, Y = np.meshgrid(X, Y)
    
    C = np.around(C)

    plt.pcolormesh(X, Y, C)

    plt.show()
'''




'''Simple tkinter widgets usable by larger widget assemblies.
'''

import os
import numpy as np
import threading

import tkinter as tk
from tkinter import filedialog

from tk_steroids.elements import (Tabs,
        ButtonsFrame,
        Listbox,
        )
from tk_steroids.dialogs import TickSelect
from gonioanalysis.droso import SpecimenGroups



def select_specimens(core, command, with_rois=None, with_movements=None, with_correction=None,
        command_args=[], execute_callable_args=True, breaking_args=[()],
        return_manalysers=False):
    '''
    Opens a specimen selection window and after ok runs command using
    selected specimens list as the only input argument.

    command                 Command to close after fly selection
    with_rois               List specimens with ROIs selected if True
    with_movements          List specimens with movements measured if True
    command_args            A list of arguments passed to the command
    execute_callable_args   Callable command_args will get executed and return
                                value is used instead
    breaking_args           If command_args callable return value in this list,
                                halt the command
    return_manalysers       Instead of passing the list of the specimen names as the first
                            argument to the command, already construct MAnalyser objects and pass those
    '''
    parsed_args = []
    for arg in command_args:
        if execute_callable_args and callable(arg):
            result = arg()
            if result in breaking_args:
                # Halting
                return None
            parsed_args.append(result)
        else:
            parsed_args.append(arg)


    top = tk.Toplevel()
    top.title('Select specimens')
    top.grid_columnconfigure(1, weight=1)
    top.grid_rowconfigure(1, weight=1)


    if with_rois or with_movements or with_correction:
        notify_string = 'Listing specimens with '
        notify_string += ' and '.join([string for onoff, string in zip([with_rois, with_movements, with_correction],
            ['ROIs', 'movements', 'correction']) if onoff ])
        tk.Label(top, text=notify_string).grid(row=0, column=1)

    specimens = core.list_specimens(with_rois=with_rois, with_movements=with_movements, with_correction=with_correction) 
    
    groups = list(SpecimenGroups().groups.keys())

    if return_manalysers:
        # This is quite wierd what is going on here
        def commandx(specimens, *args, **kwargs):
            manalysers = core.get_manalysers(specimens)
            return command(manalysers, *args, **kwargs)
    else:
        commandx = command

    tabs = Tabs(top, ['Specimens', 'Groups'])

    for tab, selectable in zip(tabs.tabs, [specimens, groups]):
        selector = TickSelect(tab, selectable, commandx, callback_args=parsed_args)
        selector.grid(sticky='NSEW', row=1, column=1)
    
        tk.Button(selector, text='Close', command=top.destroy).grid(row=2, column=1)
    
    tabs.grid(row=1, column=1,sticky='NSEW')


def select_specimen_groups(core, command):
    '''
    command gets the following dictionary
        {'group1_name': [manalyser1_object, ...], ...}
    '''
    top = tk.Toplevel()
    top.title('Select specimen groups')
    top.grid_columnconfigure(0, weight=1)
    top.grid_rowconfigure(1, weight=1)

    
    gm = SpecimenGroups()
    gm.load_groups()

    def commandx(group_names):
        grouped = {}
        for group_name in group_names:
            print(gm.groups[group_name])
            manalysers = [core.get_manalyser(specimen) for specimen in gm.groups[group_name]]
            grouped[group_name] = manalysers
        command(grouped)

    selector = TickSelect(top, list(gm.groups.keys()), commandx)
    selector.grid(sticky='NSEW')

    tk.Button(selector, text='Close', command=top.destroy).grid(row=2, column=1)



class ImagefolderMultisel(tk.Frame):
    '''
    Widget to select image folders from the specimens
    
    Attributes
    ----------
    core
    specimens_listbox
    imagefolders_listbox
    buttons_frame
    '''
    
    def __init__(self, tk_parent, core, callback, **kwargs):
        '''
        *kwargs to core.list_specimens
        '''
        tk.Frame.__init__(self, tk_parent)
        
        self.tk_parent = tk_parent
        self.core = core
        self.callback = callback
        self._separator = ';'

        specimens = core.list_specimens(**kwargs) 
        self.specimens_listbox = Listbox(self, specimens, self.on_specimen_selection)
        self.specimens_listbox.grid(row=0, column=0, sticky='NSWE')
        
        self.imagefolders_listbox = Listbox(self, [''], None)
        self.imagefolders_listbox.grid(row=0, column=1, sticky='NSWE')

        self.buttons_frame = ButtonsFrame(self,
                button_names=['Add', 'Add all', 'Remove', 'Ok'], horizontal=False,
                button_commands=[self.on_add_press, self.on_add_all, self.on_remove_press, self.on_ok])
        self.buttons_frame.grid(row=0, column=2)

        self.selected_listbox = Listbox(self, [], None)
        self.selected_listbox.grid(row=0, column=3, sticky='NSWE')
        
        for i in [0, 1, 3]:
            self.grid_columnconfigure(i, weight=1)
        self.grid_rowconfigure(0, weight=1)
    
    def on_specimen_selection(self, name):
        analyser = self.core.get_manalyser(name)
        image_folders = analyser.list_imagefolders()
        self.imagefolders_listbox.set_selections(image_folders)
        
        # Bugfix: Prevent adding image_folder from the previous specimen when
        # a new specimen has been selected but no change on the image folders
        # Requires yet unreleased tk_steroids > v0.7.1
        try:
            self.imagefolders_listbox.current = None
        except NameError:
            # tk_steroids before the current setter: 
            # NameError: name 'setter' is not defined. Did you mean: 'setattr'?
            # Let the bug be...
            pass


    def on_add_press(self, image_folder=None):
        if image_folder is None:
            image_folder = self.imagefolders_listbox.current

        if image_folder:

            sel = self.specimens_listbox.current + self._separator + image_folder
            
            selections = self.selected_listbox.selections + [sel]
            self.selected_listbox.set_selections(selections)
    
    def on_add_all(self):
        for image_folder in self.imagefolders_listbox.selections:
            self.on_add_press(image_folder)

    def on_remove_press(self):
        to_remove = self.selected_listbox.current
        if to_remove:
            selections = self.selected_listbox.selections
            selections.remove(to_remove)
            
            self.selected_listbox.set_selections(selections)


    def on_ok(self):
        image_folders = {}
        for z in self.selected_listbox.selections:
            s, i = z.split(self._separator)
            if s not in image_folders:
                image_folders[s] = []
            image_folders[s].append(i)

        self.callback(image_folders)
        self.tk_parent.destroy()




class WaitingFrame(tk.Frame):

    def __init__(self, tk_master, text):
        tk.Frame.__init__(self, tk_master)
        tk.Label(self, text=text).grid()


class WaitingWindow():
    '''
    Spans a new tkinter root window so use this only
    if there's no tkinter root beforehad.
    '''
    def __init__(self, title, text):
        
        self.root = tk.Tk()
        self.root.title(title)
        WaitingFrame(self.root, text).grid()
        self.root.update()

    def close(self):
        self.root.destroy()


class RepetitionSelector(tk.Frame):

    def __init__(self, tk_master, RecordingPlotter, core, update_command):
        '''
        
        update_command      Callable that updates the plots, no input arguments
        '''

        self.update_command = update_command

        tk.Frame.__init__(self, tk_master)
        self.core = core

        self.plotter = RecordingPlotter
            
        self.text = tk.StringVar()
        self.infotext = tk.Label(self, textvariable = self.text)
        self.infotext.grid(row=0, column=0)

        self.all = tk.Button(self, text='Show all', command=lambda: self.move_selection(None))
        self.all.grid(row=0, column=1)

        self.previous = tk.Button(self, text='Previous', command=lambda: self.move_selection(-1))
        self.previous.grid(row=0, column=2)

        self.next = tk.Button(self, text='Next', command=lambda: self.move_selection(1))
        self.next.grid(row=0, column=3)

        self.mark_bad = tk.Button(self, text='Mark bad', command=self.mark_bad)
        self.mark_bad.grid(row=0, column=4)

    def mark_bad(self):
        im_folder = self.core.selected_recording
        
        if self.plotter.i_repeat == None:
            pass
        else:
            self.core.analyser.mark_bad(
                    im_folder, self.plotter.i_repeat,
                    i_is_relative=True)
        
        self.move_selection(0)


    def move_selection(self, direction):
        '''
        None        sets plotter to show all repeats
        1 or -1     Move to next/previous repetition
        '''

        if direction == None:
            self.plotter.i_repeat = None
        else:
            if self.plotter.i_repeat == None:
                self.plotter.i_repeat = 0
            else:
                self.plotter.i_repeat += direction

            if self.plotter.i_repeat < 0:
                self.plotter.i_repeat = 0
            elif self.plotter.i_repeat >= self.plotter.N_repeats:
                self.plotter.i_repeat = self.plotter.N_repeats -1
        
        self.update_text()
        
        print(self.plotter.i_repeat)

        if self.update_command:
            self.update_command()

    def update_text(self):
        if self.plotter.i_repeat is None:
            isel = None
        else:
            isel = self.plotter.i_repeat + 1
        self.text.set("{}/{}".format(isel, self.plotter.N_repeats))



class SetPlotlimits(tk.Frame):
    '''Configure matplotlib XY plot options
    
    Attributes
    ----------
    parent : tk.Frame
        The CanvasPlotter object from the tk-steroids package.
    '''

    def __init__(self, tk_parent, ax):
        tk.Frame.__init__(self, tk_parent)
        
        self.ax = ax
        self.parent = tk_parent

        self.inputs = []
        for icol, limit in enumerate(['x0', 'x1', 'y0', 'y1']):
            label = tk.Label(self, text=limit)
            label.grid(row=0, column=icol*2)
            text = tk.Text(self, height=1, width=8)
            text.grid(row=0, column=icol*2+1)
            self.inputs.append(text)

        self.apply_button = tk.Button(
                self, text='Apply', command=self.apply)
        self.apply_button.grid(row=0, column=icol*2+2)
        
        self.apply_button = tk.Button(
                self, text='Clear', command=self.clear)
        self.apply_button.grid(row=0, column=icol*2+3)
        
        self.toolbar_button = tk.Button(
                self, text='Show controls', command=self.toggle_toolbar)
        self.toolbar_button.grid(row=0, column=icol*2+4)
        

    def apply(self, values=None):

        if values is None:
            values = []
            for text in self.inputs:
                value = text.get('1.0', 'end')

                try:
                    value = float(value)
                except:
                    value = None
                
                if value != 0 and not value:
                    value = None

                values.append(value)
        
        x0,x1,y0,y1 = values
        self.ax.set_xlim(x0, x1)
        self.ax.set_ylim(y0, y1)
        
        self.parent.update()

    def clear(self):
        for text in self.inputs:
            text.delete('1.0', 'end')
        
        self.ax.autoscale()
        self.parent.update()

    def toggle_toolbar(self):
        newstate = not self.parent._toolbar_visible
        self.parent.set_toolbar_visibility(newstate)
        if newstate:
            self.toolbar_button.config(text='Hide controls')
        else:
            self.toolbar_button.config(text='Show controls')




import os
import sys
import shutil

def main():
    workdir = os.getcwd()
    pdir = os.path.dirname(sys.executable)

    shutil.copytree(os.path.join(pdir, 'tcl'), 'lib')
    os.makedirs('pynsist_pkgs', exist_ok=True)

    copyfiles = [os.path.join(pdir, 'DLLs', fn) for fn in ['_tkinter.pyd', 'tcl86t.dll', 'tk86t.dll']]
    copyfiles.append(os.path.join(pdir, 'libs', '_tkinter.lib'))

    for origfile in copyfiles:
        newfile = os.path.join('pynsist_pkgs', os.path.basename(origfile))
        shutil.copy(origfile, newfile)

if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# Version number to __version__ variable
exec(open("gonioanalysis/version.py").read())

install_requires = [
        'numpy',
        'scipy',
        'tifffile',
        'matplotlib',
        'tk-steroids>=0.8.1',
        'roimarker>=0.2.1',
        'movemeter>=0.7.2',
        'python-biosystfiles',
        'devjoni-hosguibase>=0.0.6',
        'Panda3D>=1.10.0'
        ]


setuptools.setup(
    name="gonio-analysis",
    version=__version__,
    author="Joni Kemppainen",
    author_email="joni.kemppainen@windowslive.com",
    description="Spatial motion analysis program",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/musaprog/gonio-analysis",
    packages=setuptools.find_packages(),
    install_requires=install_requires, 
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    # Used language features that require Python 3.6 or newer
    #   - fstrings
    #
    #
    # Python 3.8 final version supporting Windows 7
    # Python 3.4 final version supporting Windows XP
    python_requires='>=3.6',

    include_package_data=True,
    package_data={'gonioanalysis.progdata': ['*.egg']}
)



'''
Rhabdomere orientation
'''
import os
import json

import matplotlib.pyplot as plt

from roimarker import Marker

from gonioanalysis.drosom.analysing import MAnalyser
from gonioanalysis.directories import PROCESSING_TEMPDIR


class OAnalyser(MAnalyser):
    '''
    Rhabdomere orientation analyser.

    Inherits from MAnalyser though most of it's methods
    have no meaning.
    
    Measure movement opens Marker to draw lines/arrows

    '''

    def __init__(self, *args, **kwargs):


        super().__init__(*args, **kwargs)
        
        self._movements_skelefn = self._movements_skelefn.replace('movements_', 'orientation_')
        self.active_analysis = ''


    def measure_movement(self, eye, *args, **kwargs):
        '''
        The measure movement method overridden to meausure the (rhabdomere)
        orientation.

        In the end calls self.load_analysed_movements in order to
        match the MAnalyser behaviour.
        '''

        self.movements = {}

        images = []
        rois = []
            
        for angle in self.stacks:
            
            roi = self.ROIs[eye].get(angle, None)
            
            if roi is not None:
                images.append(self.stacks[angle][0][0])
                

                extended_roi = [roi[0]-roi[2]/2, roi[1]-roi[3]/2, 2*roi[2], 2*roi[3]]

                rois.append(extended_roi)

        fig, ax = plt.subplots(num='Draw arrows for the {} eye'.format(eye))
        marker = Marker(fig, ax, images, self._movements_savefn.format(eye),
                relative_fns_from=os.path.join(self.data_path, self.folder),
                drop_imagefn=True,
                selection_type='arrow',
                crops=rois,
                callback_on_exit=lambda eye=eye: self._hotedit_marker_output(eye))

        marker.run()

        print('Marker should run now')


    def _hotedit_marker_output(self, eye):
        '''
        Edits Marker output to be Movemeter like output.
        '''

        with open(self._movements_savefn.format(eye), 'r') as fp:
            marker_data = json.load(fp)

        edited_data = {}

        for image_folder, arrows in marker_data.items():
            
            repeats = []

            for arrow in arrows:

                if len(arrow) == 2:
                    # Already edited, arrow is a dict we earlier made below
                    repeats.append( arrow )
                else:
                    # Needs hotediting
                    x1, y1, x2, y2 = arrow
                    repeats.append( {'x': [0, x1-x2], 'y': [0, y1-y2]} )

            # drop pos prefix [3:]
            if repeats != []:
                edited_data[image_folder[3:]] = repeats
       

        with open(self._movements_savefn.format(eye), 'w') as fp:
            json.dump(edited_data, fp)

   
    
    def is_measured(self):
        fns = [self._movements_savefn.format(eye) for eye in self.eyes]
        return all([os.path.exists(fn) for fn in fns])






import io
import sys
import threading

import tkinter as tk
from tk_steroids.elements import BufferShower

class MeasurementWindow:
    '''
    Sets the manalyser to do the movement measurements and
    opens a window showing the progress.

    Because this sets stdout to a StringIO temporalily, it's best
    to run this in a subprocess.
    '''

    def __init__(self, tk_root, thread_targets, title='', callback_on_exit=None):
        '''
        tk_root         Tkinter root object, needed for scheduling events with after-method
        thread_targets  Callables
        '''
        self.root = tk_root
        self.title = title
        self.callback_on_exit = callback_on_exit

        self.i_target = -1
        self.thread_targets = thread_targets
        self.processes = []

        self.all_run = False
        self.exit = False
        self.run()


    def run(self):
        self.top = tk.Toplevel()
        self.top.title(self.title)

        self.oldout = sys.stdout
        sys.stdout = io.StringIO()

        BufferShower(self.top, sys.stdout).grid()
        self.cancel_button = tk.Button(self.top, text='Cancel', command=self.on_cancel)
        self.cancel_button.grid()
        
        self.check_finished()


    def _run_next_target(self):
        '''
        Set next manalyser to work or return False if none left.
        '''
        self.i_target += 1
        
        if self.i_target == len(self.thread_targets):
            return False

        self.stop_event = threading.Event()
        
        p = threading.Thread(target=self.thread_targets[self.i_target],
                args=[self.stop_event,])

        p.start()
        self.processes.append(p)
    
        return True
    

    def alives(self):
        return [process.is_alive() for process in self.processes]


    def check_finished(self):
        '''
        Check if all the threads are finished and if there are more targets to run.
        Reschedule every 1000 ms.
        '''
        if not self.exit:
            if not any(self.alives()):
                targets_left = self._run_next_target()
            else:
                targets_left = True

            if targets_left:
                self.root.after(1000, self.check_finished)
            else:
                self.cancel_button.config(text='Ok')
                self.all_run = True


    def on_cancel(self):
        self.exit = True
        
        # Calcelled, only send thread signal to stop
        self.stop_event.set()
        
        if self.all_run:
            self.callback_on_exit()

        if not all(self.alives()):
            self.top.destroy()
            sys.stdout = self.oldout
        else:
            self.root.after(1000, self.on_cancel)


def main():
    '''
    Read data directory and specimen name from sys.argv
    and run movement measurement for that manalyser.
    '''
    pass

if __name__ == main():
    main()




import os
import json

import numpy as np
import tifffile

from gonioanalysis.drosom.analysing import MAnalyser



class TAnalyser(MAnalyser):
    '''
    Transmittance analyser.

    Analyses the ROI's light throughput while the ROI tracks
    its moving feature (ie. using MAnalyser motion analysis results)
    
    Mean (average) pixel value of the ROI is quantified.
    '''

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        
        self._real_movements_savefn = self._movements_savefn

        self._movements_skelefn = self._movements_skelefn.replace('movements_', 'transmittance_')
        self.active_analysis = ''


    def measure_movement(self, eye, *args, **kwargs):
        '''
        Analyse transmittance/brightness by calculating mean (average)
        pixel value of the ROI in its time locations, and save results.
        '''
        self.movements = {}
        
        intensities = {}

        manalyser = MAnalyser(self.data_path, self.folder)
        manalyser.active_analysis = self.active_analysis

        for i, angle in enumerate(self.stacks):
            print('  Image folder {}/{}'.format(i+1, len(self.stacks)))
            
            roi = self.ROIs[eye].get(angle, None)

            if roi is not None:
                
                images = self.stacks[angle]
                
                intensities[angle] = []

                for i_repeat, repeat in enumerate(images):
                    ints = []
                    
                    try:
                        _roi = manalyser.get_moving_ROIs(eye, angle, i_repeat)
                    except AttributeError:
                        _roi = None
                    
                    i_frame = 0

                    for fn in repeat:
                        
                        tiff = tifffile.TiffFile(fn)
                        
                        for i_page in range(len(tiff.pages)):
                            
                            images = tiff.asarray(key=i_page)
                            if len(images.shape) == 2:
                                images = [images]
                            
                            for image in images:
                                if _roi is None:
                                    x,y,w,h = roi
                                else:
                                    try:
                                        x,y,w,h = [int(round(z)) for z in _roi[i_frame]]
                                    except IndexError:
                                        # No ROI movement, skip
                                        break
                                
                                intensity = np.mean(image[y:y+h,x:x+w])
                                ints.append(intensity)
                                
                                i_frame += 1
                                print("fn {}: {}/{}".format(os.path.basename(fn), i_frame+1, len(tiff.pages)))

                    # X component carries the brightness change information
                    # Y component is set to zero for compability (in future
                    #   remove to save disk?)
                    intensities[angle].append({'x': ints, 'y':[0]*len(ints)})


        self.movements[eye] = intensities

        # Save movements
        with open(self._movements_savefn.format(eye), 'w') as fp:
            json.dump(intensities, fp)

    

    def is_measured(self):
        fns = [self._movements_savefn.format(eye) for eye in self.eyes]
        return all([os.path.exists(fn) for fn in fns])


    def load_analysed_movements(self, *args, **kwargs):
        super().load_analysed_movements(*args, **kwargs)

        # Nullify the Y component so that X carries the brightness change data.
        # Otherwise, at get_magnitude traces, when calculating displacement
        # distance, the data gets multiplied by sqrt(2).
        #
        # The same change is done in TAnalyser.measure_movement but this also
        # fixes previously analysed data without reanalysing.
        for eye in self.movements:
            for image_folder, data in self.movements[eye].items():
                for i_repeat in range(len(data)):
                    data[i_repeat]['y'] = np.zeros(len(data[i_repeat]['x']))


    # Here the original MAnalyser method modified so that the presentation
    # makes more sense with light transmittance analysis 
    #
    # 1) The first data point is often "polluted" because the stimulus LED
    #   goes on within the first frame of imaging (thus first frame darker)
    #   -> bring the 1st to the mean of 2nd and 3rd
    # 2) Shift the trace in Y to always start from zero
    #
    def get_magnitude_traces(self, *args, **kwargs):
        
        magnitude_traces = super().get_magnitude_traces(*args, **kwargs)

        for image_folder in magnitude_traces:

            # 1) "Recompute" the first data point
            for trace in magnitude_traces[image_folder]:
                trace[0] = np.mean((trace[1], trace[2]))

            # 2) Shift whole trace
            magnitude_traces[image_folder] = np.array([t-t[0] for t in magnitude_traces[image_folder]])

        return magnitude_traces



class RelativeTAnalyser(TAnalyser):
    '''Transmittance analysis but report relative changes.

    Changes are relative to the 1st frame photon count of each repeat (the
    1st frame photon count is corrected to be the mean of 2nd and 3rd frames).
    
    Uses TAnalyser measurements and vice versa so measuring either with
    TAnalyser or RelativeTAnalyser makes both measured.
    '''

    def load_analysed_movements(self, *args, **kwargs):
        super().load_analysed_movements(*args, **kwargs)

        for eye in self.movements:
            for image_folder, data in self.movements[eye].items():

                for i_repeat in range(len(data)):
                    for key in ['x', 'y']:
                        start_val = np.mean((data[i_repeat][key][1], data[i_repeat][key][2]))
                        data[i_repeat][key][0] = start_val
                        if start_val == 0:
                            continue

                        # Add on top of a big number (1000) so that negative
                        # changes won't die in the squaring when getting mag
                        # traces.
                        data[i_repeat][key] = 1000+(
                                (np.array(data[i_repeat][key]) - start_val) / start_val)



__version__ = "0.8.8"


# Need for easily checking the scipy version because of API changes
# For example, change in KDTree.query arguments: i_jobs -> workers
from scipy import __version__ as used_scipy_version
used_scipy_version = tuple([int(v) for v in used_scipy_version.split('.')])
